@@ -21,6 +21,7 @@ import datetime
 import socket
 import operator
 import re
+import sys
 from six.moves import range
 from bitarray import bitarray
 
@@ -37,7 +38,7 @@ from impala._thrift_api import (
     TGetTablesReq, TGetFunctionsReq, TGetOperationStatusReq, TOperationState,
     TCancelOperationReq, TCloseOperationReq, TGetLogReq, TProtocolVersion,
     TGetRuntimeProfileReq, TGetExecSummaryReq, ImpalaHiveServer2Service,
-    TExecStats, ThriftClient)
+    TExecStats, ThriftClient, TApplicationException)
 
 
 log = get_logger_and_init_null(__name__)
@@ -76,7 +77,8 @@ class HiveServer2Connection(Connection):
         # PEP 249
         raise NotSupportedError
 
-    def cursor(self, user=None, configuration=None, convert_types=True, dictify=False):
+    def cursor(self, user=None, configuration=None, convert_types=True,
+               dictify=False, fetch_error=True):
         """Get a cursor from the HiveServer2 (HS2) connection.
 
         Parameters
@@ -87,7 +89,25 @@ class HiveServer2Connection(Connection):
         convert_types : bool, optional
             When `False`, timestamps and decimal values will not be converted
             to Python `datetime` and `Decimal` values. (These conversions are
-            expensive.)
+            expensive.) Only applies when using HS2 protocol versions > 6.
+        dictify : bool, optional
+            When `True` cursor will return key value pairs instead of rows.
+        batch_cursor : bool, optional
+            When `True` cursor will return CBatches directly rather than rows.
+        fetch_error : bool, optional
+            In versions of impala prior to 2.7.0, when an operation fails and
+            the impalad returns an error state, the error message is not always
+            returned. In these cases the error message can be retrieved by a
+            subsequent fetch rpc call but this has a side effect of invalidating
+            the query handle and causing any further operations against it to
+            fail. e.g. calling log() or profile().
+
+            When set to `True` impyla will attempt to fetch the error message.
+            When set to `False`, this flag will cause impyla not to attempt to
+            fetch the message with a fetch call . In this case the query
+            handle remains valid and impyla will raise an exception with a
+            message of "Operation is in ERROR_STATE".
+            The Default option is `True`.
 
         Returns
         -------
@@ -111,7 +131,8 @@ class HiveServer2Connection(Connection):
 
         cursor_class = HiveServer2DictCursor if dictify else HiveServer2Cursor
 
-        cursor = cursor_class(session, convert_types=convert_types)
+        cursor = cursor_class(session, convert_types=convert_types,
+                              fetch_error=fetch_error)
 
         if self.default_db is not None:
             log.info('Using database %s as default', self.default_db)
@@ -128,9 +149,10 @@ class HiveServer2Cursor(Cursor):
     # HiveServer2Cursor objects are associated with a Session
     # they are instantiated with alive session_handles
 
-    def __init__(self, session, convert_types=True):
+    def __init__(self, session, convert_types=True, fetch_error=True):
         self.session = session
         self.convert_types = convert_types
+        self.fetch_error = fetch_error
 
         self._last_operation = None
 
@@ -143,6 +165,18 @@ class HiveServer2Cursor(Cursor):
         self._description = None
         self._rowcount = -1
 
+        self._closed = False
+
+    def __del__(self):
+        try:
+           self.close_operation()
+        except Exception:
+            pass
+        try:
+           self.session.close()
+        except Exception:
+            pass
+
     @property
     def description(self):
         # PEP 249
@@ -189,16 +223,36 @@ class HiveServer2Cursor(Cursor):
 
     def close(self):
         # PEP 249
+        if self._closed:
+            return
         # If an operation is active and isn't closed before the session is
         # closed, then the server will cancel the operation upon closing
         # the session. Cancellation could be problematic for some DDL
         # operations. This avoids requiring the user to call the non-PEP 249
         # close_operation().
-        self.close_operation()
+        exc_info = None
+        try:
+            self.close_operation()
+        except Exception:
+            exc_info = sys.exc_info()
 
         log.debug('Closing HiveServer2Cursor')
+        try:
+            self.session.close()
+        except Exception:
+            # If we encountered an error when closing the session
+            # then print operation close exception to logs and
+            # raise the session close exception
+            if exc_info:
+                log.error('Failure encountered closing last operation.',
+                                                        exc_info=exc_info)
+            raise
 
-        self.session.close()
+        self._closed = True
+        # If there was an error when closing last operation then
+        # raise exception
+        if exc_info:
+            six.reraise(*exc_info)
 
     def cancel_operation(self):
         if self._last_operation_active:
@@ -310,14 +364,26 @@ class HiveServer2Cursor(Cursor):
         self._debug_log_state()
 
     def _wait_to_finish(self):
+        # Prior to IMPALA-1633 GetOperationStatus does not populate errorMessage
+        # in case of failure. If not populated, queries that return results
+        # can get a failure description with a further call to FetchResults rpc.
         loop_start = time.time()
         while True:
-            operation_state = self._last_operation.get_status()
+            req = TGetOperationStatusReq(operationHandle=self._last_operation.handle)
+            resp = self._last_operation._rpc('GetOperationStatus', req)
+            operation_state = TOperationState._VALUES_TO_NAMES[resp.operationState]
 
             log.debug('_wait_to_finish: waited %s seconds so far',
                       time.time() - loop_start)
             if self._op_state_is_error(operation_state):
-                raise OperationalError("Operation is in ERROR_STATE")
+                if resp.errorMessage:
+                    raise OperationalError(resp.errorMessage)
+                else:
+                    if self.fetch_error and self.has_result_set:
+                        self._last_operation_active=False
+                        self._last_operation.fetch()
+                    else:
+                        raise OperationalError("Operation is in ERROR_STATE")
             if not self._op_state_is_executing(operation_state):
                 break
             time.sleep(self._get_sleep_interval(loop_start))
@@ -378,6 +444,35 @@ class HiveServer2Cursor(Cursor):
         except StopIteration:
             return None
 
+
+    def fetchcbatch(self):
+        '''Return a CBatch object of any data currently in the buffer or
+           if no data currently in buffer then fetch a batch'''
+        if not self._last_operation.is_columnar:
+            raise NotSupportedError("Server does not support columnar "
+                                    "fetching")
+        if not self.has_result_set:
+            raise ProgrammingError(
+                "Trying to fetch results on an operation with no results.")
+        if len(self._buffer) > 0:
+            log.debug('fetchcbatch: buffer has data in. Returning it and wiping buffer')
+            batch = self._buffer
+            self._buffer = Batch()
+            return batch
+        elif self._last_operation_active:
+            log.debug('fetchcbatch: buffer empty and op is active => fetching '
+                      'more data')
+            batch = (self._last_operation.fetch(
+                         self.description,
+                         self.buffersize,
+                         convert_types=self.convert_types))
+            if len(batch) == 0:
+               return None
+            return batch
+        else:
+           return None
+
+
     def fetchmany(self, size=None):
         # PEP 249
         self._wait_to_finish()
@@ -447,7 +542,8 @@ class HiveServer2Cursor(Cursor):
             log.debug('__next__: buffer empty and op is active => fetching '
                       'more data')
             self._buffer = self._last_operation.fetch(self.description,
-                                                      self.buffersize)
+                                                      self.buffersize,
+                                                      convert_types=self.convert_types)
             if len(self._buffer) == 0:
                 log.debug('__next__: no more data to fetch')
                 raise StopIteration
@@ -544,11 +640,15 @@ class HiveServer2Cursor(Cursor):
 
 
 class HiveServer2DictCursor(HiveServer2Cursor):
+
     """The cursor that returns each element as a dictionary"""
     def execute(self, operation, parameters=None, configuration=None):
         super(self.__class__, self).execute(operation, parameters,
                                             configuration)
-        self.fields = [d[0] for d in self.description]
+        if self.description is not None:
+            self.fields = [d[0] for d in self.description]
+        else:
+            self.fields = None
 
     def __next__(self):
         record = super(self.__class__, self).__next__()
@@ -708,18 +808,25 @@ class Column(object):
         self.data_type = data_type
         self.values = values
         self.nulls = nulls
+        self.rows_left = len(self.values)
+        self.num_rows = self.rows_left
 
     def __len__(self):
-        return len(self.values)
+        return self.rows_left
 
     def __str__(self):
         return 'Column(type={0}, values={1}, nulls={2})'.format(
             self.data_type, self.values, self.nulls)
 
     def pop(self):
-        isnull = self.nulls.pop(0)
-        value = self.values.pop(0)
-        return None if isnull else value
+        if self.rows_left < 1:
+            raise StopIteration
+        pos = self.num_rows-self.rows_left
+        self.rows_left -= 1
+        if self.nulls[pos]:
+           return None
+        value = self.values[pos]
+        return value
 
 
 class CBatch(Batch):
@@ -730,6 +837,7 @@ class CBatch(Batch):
                  for (i, col) in enumerate(trowset.columns)]
         num_cols = len(tcols)
         num_rows = len(tcols[0].values)
+
         log.debug('CBatch: input TRowSet num_cols=%s num_rows=%s tcols=%s',
                   num_cols, num_rows, tcols)
 
@@ -1004,9 +1112,22 @@ class Operation(ThriftRPC):
         resp = self._rpc('GetOperationStatus', req)
         return TOperationState._VALUES_TO_NAMES[resp.operationState]
 
-    def get_log(self):
-        req = TGetLogReq(operationHandle=self.handle)
-        return self._rpc('GetLog', req).log
+    def get_log(self, max_rows=1024, orientation=TFetchOrientation.FETCH_NEXT):
+        try:
+            req = TGetLogReq(operationHandle=self.handle)
+            log = self._rpc('GetLog', req).log
+        except TApplicationException as e: # raised if Hive is used
+            if not e.type == TApplicationException.UNKNOWN_METHOD:
+                raise
+            req = TFetchResultsReq(operationHandle=self.handle,
+                                   orientation=orientation,
+                                   maxRows=max_rows,
+                                   fetchType=1)
+            resp = self._rpc('FetchResults', req)
+            schema = [('Log', 'STRING', None, None, None, None, None)]
+            log = self._wrap_results(resp.results, schema, convert_types=True)
+            log = '\n'.join(l[0] for l in log)
+        return log
 
     def cancel(self):
         req = TCancelOperationReq(operationHandle=self.handle)
